# -*- coding: utf-8 -*-
# Corpuscula project: Corpus Dictionary
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
A corpus dictionary to use for further morphology processing. It contains
a bunch of useful statictics and the generator of LEMMA by the word FORM and
POS tag.
"""
from operator import itemgetter
import pickle
import sys

from corpuscula.conllu import Conllu
from corpuscula.corpus_utils import _AbstractCorpus
from corpuscula.utils import LOG_FILE, find_affixes, print_progress


class CorpusDict:
    """Arrays of information extracted from labeled corpus(es).
    May be used for a lot of text processing tasks"""

    def __init__(self, restore_from=None, corpus=None, format='conllu',
                 backup_to=None, cnt_thresh=20, ambiguity_thresh=1.,
                 log_file=LOG_FILE):
        """
        :param restore_from: path to backup file to load from
        :type restore_from: str
        :param corpus: path to file in known format or list of already parsed
                       sentences
        :type corpus: str|list(dict())
        :param format: known format: 'conllu'|'conllu_parsed'
        :type format: str
        :param backup_to: path to backup file to save to
        :type backup_to: str
        :param cnt_thresh: param for ``fit()``
        :param ambiguity_thresh: param for ``fit()``
        :param log_file: stream for info messages
        :type file

        NB: If both *restore_from* and *corpus* are not None then information
        extracted from *corpus* will be added to state restored from backup"""
        self._cnt_thresh       = cnt_thresh
        self._ambiguity_thresh = ambiguity_thresh

        self._wforms    = []  # [wform]
        self._lemmata   = []  # [lemma]
        self._tags      = []  # [tag]
        self._feats     = []  # [feat]
        self._feat_vals = []  # [[val]]

        self._wforms_id    = {}  # {wform: id}
        self._lemmata_id   = {}  # {lemma: id}
        self._tags_id      = {}  # {tag: id}
        self._feats_id     = {}  # {feat: id}
        self._feat_vals_id = []  # [{val: id}]

        self._tags_freq = []       # [(tag, cnt, freq)], sorted by freq
        self._feats_freq = {}      # {tag: [(feat, cnt, freq)]}, sorted by freq
        self._feat_vals_freq = {}  # {tag: {feat: [(val, cnt, freq)]}},
                                   # sorted by freq

        self._wform_tag_cnts  = {}  # {wform_id: {tag_id: {lemma_id: cnt}}}
        self._wform_feat_cnts = {}
            # {wform_id: {tag_id: {lemma_id: {feat_id: {val_id: cnt}}}}}
            #                                 '--- may be empty ---'
        self._lemma_tag_cnts  = {}  # {lemma_id: {tag_id: cnt}}
        self._lemma_feat_cnts = {}
            # {lemma_id: {tag_id: {feat_id: {val_id: cnt}}}}
            #                      '--- may be empty ---'

        self._most_probable_tags        = {}    # {wform_id: tag_id}
        self._most_common_tag           = None  # tag
        self._common_endings            = {}
            # {tag_id: {wform_ed: [(lemma_ed, cnt)]})
            # [(lemma_ed, cnt)] sorted by cnt
        self._capitalized_tags          = set() # set(tag_id)
        self._tag_feats                 = []    # [set(feat_id)]
        self._most_probable_wform_feats = {}    # {wform_id: {feat_id: val_id}}
        self._most_probable_lemma_feats = {}    # {lemma_id: {feat_id: val_id}}

        if restore_from:
            self.restore_from(restore_from)
        if corpus:
            self.parse(corpus, format=format, log_file=log_file)
        if backup_to:
            self.backup_to(backup_to)
        if corpus:
            self.fit(log_file=log_file)

    def backup(self):
        """Get current state"""
        return {'_cnt_thresh'      : self._cnt_thresh      ,
                '_ambiguity_thresh': self._ambiguity_thresh,
                '_wforms_id'       : self._wforms_id       ,
                '_lemmata_id'      : self._lemmata_id      ,
                '_tags_id'         : self._tags_id         ,
                '_feats_id'        : self._feats_id        ,
                '_feat_vals_id'    : self._feat_vals_id    ,
                '_wform_tag_cnts'  : self._wform_tag_cnts  ,
                '_wform_feat_cnts' : self._wform_feat_cnts ,
                '_lemma_tag_cnts'  : self._lemma_tag_cnts  ,
                '_lemma_feat_cnts' : self._lemma_feat_cnts }

    def backup_to(self, file_path):
        """Store current state to file"""
        with open(file_path, 'wb') as f:
            pickle.dump(self.backup(), f, 2)

    def restore(self, o, log_file=LOG_FILE):
        """Restore current state from backup object"""
        (self._cnt_thresh      ,
         self._ambiguity_thresh,
         self._wforms_id       ,
         self._lemmata_id      ,
         self._tags_id         ,
         self._feats_id        ,
         self._feat_vals_id    ,
         self._wform_tag_cnts  ,
         self._wform_feat_cnts ,
         self._lemma_tag_cnts  ,
         self._lemma_feat_cnts ) = [o[x] for x in ['_cnt_thresh'      ,
                                                   '_ambiguity_thresh',
                                                   '_wforms_id'       ,
                                                   '_lemmata_id'      ,
                                                   '_tags_id'         ,
                                                   '_feats_id'        ,
                                                   '_feat_vals_id'    ,
                                                   '_wform_tag_cnts'  ,
                                                   '_wform_feat_cnts' ,
                                                   '_lemma_tag_cnts'  ,
                                                   '_lemma_feat_cnts' ]]
        self._wforms = list(k for k, _ in
            sorted(self._wforms_id.items(), key=itemgetter(1)))
        self._lemmata = list(k for k, _ in
            sorted(self._lemmata_id.items(), key=itemgetter(1)))
        self._tags = list(k for k, _ in
            sorted(self._tags_id.items(), key=itemgetter(1)))
        self._feats = list(k for k, _ in
            sorted(self._feats_id.items(), key=itemgetter(1)))
        self._feat_vals = list(
            [k for k, _ in sorted(vals.items(), key=itemgetter(1))]
                for vals in self._feat_vals_id
        )
        self.fit(log_file=log_file)

    def restore_from(self, file_path, log_file=LOG_FILE):
        """Restore current state from file"""
        with open(file_path, 'rb') as f:
            self.restore(pickle.load(f), log_file=log_file)

    def isempty(self):
        """Check if current state does not contain any information"""
        return not self._wforms

    def parse(self, corpus, format='conllu', append=False, log_file=LOG_FILE):
        """Extract useful information from a *corpus* given.

        :param corpus: path to file in known format or list of already parsed
                       sentences
        :type corpus: str|list(dict())
        :param format: known format: 'conllu'|'conllu_parsed'
        :type format: str
        :param append: if False then method will throw exception if current
                       state is not empty
        :type append: bool
        :param log_file: stream for info messages
        :type file
        """
        assert self.isempty() or append, 'Error: Current state is not ' \
            'empty. Use append=True to append next corpus'
        assert format in ['conllu', 'conllu_parsed'], \
            "Error: Invalid format '{}'".format(format)
        if isinstance(corpus, type) and issubclass(corpus,
                                                   _AbstractCorpus):
            corpus = corpus.train()
        elif format == 'conllu':
            corpus = Conllu.load(corpus, log_file=None)
        corpus_len = None if format == 'conllu' else \
                     len(corpus) if isinstance(corpus, list) else \
                     None
        progress_step = max(int(corpus_len / 60), 1000) if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100

        if log_file:
            print('Parse corpus', file=log_file)
        ntoken = 0
        nyo = 0
        sent_no = -1
        for sent_no, sentence in enumerate(corpus):
            if log_file and not sent_no % progress_check_step:
                print_progress(sent_no, end_value=corpus_len,
                               step=progress_step, file=log_file)
            if isinstance(sentence, tuple):
                sentence = sentence[0]
            for token in sentence:
                id_, wform, lemma, tag, feats = [
                    token[x] for x in ['ID', 'FORM', 'LEMMA', 'UPOS', 'FEATS']
                ]
                if wform and '-' not in id_ \
               and not any(x.isdecimal() for x in wform) \
               and not any(x.isdecimal() for x in lemma):
                    ntoken += 1
                    # if wform contain Russian "yo" letter, we'll add both two
                    # wforms instead of one: with and without "yo" with the
                    # same lemma and other attributes
                    wforms_ = [wform]
                    wform_ = wform.replace('ё', 'е').replace('Ё', 'Е')
                    if wform_ != wform:
                        wforms_.append(wform_)
                        nyo += 1
                    for wform in wforms_:
                        if lemma.islower():
                            wform = wform.lower()
                        wform_id = self._wforms_id.get(wform)
                        if wform_id is None:
                            wform_id = len(self._wforms)
                            self._wforms.append(wform)
                            self._wforms_id[wform] = wform_id
                        lemma_id = self._lemmata_id.get(lemma)
                        if lemma_id is None:
                            lemma_id = len(self._lemmata)
                            self._lemmata.append(lemma)
                            self._lemmata_id[lemma] = lemma_id
                        tag_id = self._tags_id.get(tag)
                        if tag_id is None:
                            tag_id = len(self._tags)
                            self._tags.append(tag)
                            self._tags_id[tag] = tag_id

                        self._wform_tag_cnts[wform_id][tag_id][lemma_id] = \
                            self._wform_tag_cnts.setdefault(wform_id, {}) \
                                                .setdefault(tag_id, {}) \
                                                .get(lemma_id, 0) + 1
                        self._lemma_tag_cnts[lemma_id][tag_id] = \
                            self._lemma_tag_cnts.setdefault(lemma_id, {}) \
                                                .get(tag_id, 0) + 1

                        wform_cnts_ = self._wform_feat_cnts \
                                          .setdefault(wform_id, {}) \
                                          .setdefault(tag_id, {}) \
                                          .setdefault(lemma_id, {})
                        lemma_cnts_ = self._lemma_feat_cnts \
                                          .setdefault(lemma_id, {}) \
                                          .setdefault(tag_id, {})

                        for feat, val in feats.items():
                            feat_id = self._feats_id.get(feat)
                            val_id = 1
                            if feat_id is None:
                                feat_id = len(self._feats)
                                self._feats.append(feat)
                                self._feat_vals.append(['_', val])
                                self._feats_id[feat] = feat_id
                                self._feat_vals_id.append({'_': 0,
                                                           val: val_id})
                            else:
                                vals = self._feat_vals[feat_id]
                                vals_id = self._feat_vals_id[feat_id]
                                val_id = vals_id.get(val)
                                if val_id is None:
                                    val_id = len(vals)
                                    vals.append(val)
                                    vals_id[val] = val_id
                            wform_cnts_[feat_id][val_id] = \
                                wform_cnts_.setdefault(feat_id, {}) \
                                           .get(val_id, 0) + 1
                            lemma_cnts_[feat_id][val_id] = \
                                lemma_cnts_.setdefault(feat_id, {}) \
                                           .get(val_id, 0) + 1
                elif tag is not None:
                    tag_id = self._tags_id.get(tag)
                    if tag_id is None:
                        tag_id = len(self._tags)
                        self._tags.append(tag)
                        self._tags_id[tag] = tag_id
                    for feat, val in feats.items():
                        feat_id = self._feats_id.get(feat)
                        val_id = 1
                        if feat_id is None:
                            feat_id = len(self._feats)
                            self._feats.append(feat)
                            self._feat_vals.append(['_', val])
                            self._feats_id[feat] = feat_id
                            self._feat_vals_id.append({'_': 0,
                                                       val: val_id})
                        else:
                            vals = self._feat_vals[feat_id]
                            vals_id = self._feat_vals_id[feat_id]
                            val_id = vals_id.get(val)
                            if val_id is None:
                                val_id = len(vals)
                                vals.append(val)
                                vals_id[val] = val_id

        if log_file:
            sent_no += 1
            print_progress(sent_no,
                           end_value=corpus_len if corpus_len else 0,
                           step=progress_step, file=log_file)
            print(('done: {} sentences, {} acceptable tokens '
                   '(plus {} for YO letters)')
                      .format(sent_no, ntoken, nyo),
                  file=log_file)

    def fit(self, cnt_thresh=None, ambiguity_thresh=None, log_file=LOG_FILE):
        """Gather additional statictics from corups information.

        :type cnt_thresh: int
        :type ambiguity_thresh: float
        If any wform has met in the corpus at least *cnt_thresh* times and it
        was tagged by the same label at least in (*ambiguity_thresh* * 100)%
        cases, then that label will mark as trusted for that wform.

        :param log_file: stream for info messages
        :type file
        """
        if cnt_thresh is None:
            cnt_thresh = self._cnt_thresh
        else:
            self._cnt_thresh = cnt_thresh
        if ambiguity_thresh is None:
            ambiguity_thresh = self._ambiguity_thresh
        else:
            self._ambiguity_thresh = ambiguity_thresh

        if log_file:
            print('Fit corpus dict...', end=' ', file=log_file)
            log_file.flush()

        tag_id_cnts = {}
        total_cnt = 0

        self._most_probable_tags = {}
        for wform_id, item in self._wform_tag_cnts.items():
            tag_cnts = {x: sum(y.values()) for x, y in item.items()}
            for tag_id, cnt in tag_cnts.items():
                tag_id_cnts[tag_id] = tag_id_cnts.get(tag_id, 0) + cnt
                total_cnt += cnt
            tag_id, cnt = max(tag_cnts.items(), key=itemgetter(1))
            n = sum(tag_cnts.values())
            # Don't add rare wforms to the tag dictionary
            # Only add quite unambiguous wforms
            if n >= cnt_thresh and cnt / n >= ambiguity_thresh:
                self._most_probable_tags[wform_id] = tag_id

        self._tags_freq = [(self._tags[x], y, y / total_cnt)
                               for x, y in tag_id_cnts.items()]
        self._tags_freq.sort(key=lambda x: (x[1], x[0]), reverse=True)
        #self._most_common_tag = self._tags[
        #    max(tag_id_cnts.keys(), key=lambda x: (tag_id_cnts[x], x))
        #]
        if self._tags_freq:
            self._most_common_tag = self._tags_freq[0][0] 

        self._common_endings = eds = {}
        for wform_id, item in self._wform_tag_cnts.items():
            for tag_id, item in item.items():
                for lemma_id, cnt in item.items():
                    wform = self._wforms[wform_id]
                    lemma = self._lemmata[lemma_id]
                    if wform.isalpha() and lemma.isalpha():
                        wop, wcp, wed, lop, lcp, led = \
                            find_affixes(wform, lemma, lower=True)
                        if wcp and wop == lop:
                            for i in range(len(wcp) + 1):
                                wed_ = wcp[i:] + wed
                                led_ = lcp[i:] + led
                                eds[tag_id][wed_][led_] = \
                                    eds.setdefault(tag_id, {}) \
                                       .setdefault(wed_, {}) \
                                       .get(led_, 0) + 1
        for _, items in eds.items():
            for wed, led_cnts in items.items():
                items[wed] = sorted(led_cnts.items(),
                                    # x[0] for stability
                                    key=lambda x: (x[1], x[0]), reverse=True)

        # _capitalized_tags: Tags with mostly capitalized wforms
        tag_lower_cnts, tag_upper_cnts = {}, {}
        for lemma_id, tag_cnts in self._lemma_tag_cnts.items():
            lemma = self._lemmata[lemma_id]
            if lemma.isalpha():
                islower = lemma.islower()
                for tag_id, cnt in tag_cnts.items():
                    if islower:
                        tag_lower_cnts[tag_id] = \
                            tag_lower_cnts.get(tag_id, 0) + 1
                    else:
                        tag_upper_cnts[tag_id] = \
                            tag_upper_cnts.get(tag_id, 0) + 1
        self._capitalized_tags = set(
            x for x in range(len(self._tags))
                if tag_upper_cnts.get(x, 0) > tag_lower_cnts.get(x, 0))

        tag_feats = self._tag_feats = [set() for _ in range(len(self._tags))]
        feat_val_id_cnts            = [{}    for _ in range(len(self._tags))]
        for wform_id, item in self._wform_feat_cnts.items():
            for tag_id, item in item.items():
                for lemma_id, item in item.items():
                    for feat_id, item in item.items():
                        tag_feats[tag_id].add(feat_id)
                        for val_id, cnt in item.items():
                            feat_val_id_cnts[tag_id][feat_id][val_id] = \
                                feat_val_id_cnts[tag_id].setdefault(feat_id,
                                                                    {}) \
                                                        .get(val_id, 0) + cnt

        for tag_id, item in enumerate(feat_val_id_cnts):
            tag = self._tags[tag_id]
            total_cnt = sum(x for x in item.values() for x in x.values())
            feats_freq = self._feats_freq[tag] = []
            feat_vals = self._feat_vals_freq[tag] = {}
            for feat_id, item in item.items():
                feat = self._feats[feat_id]
                vals = self._feat_vals[feat_id]
                total_cnt2 = sum(item.values())
                feats_freq.append((feat, total_cnt2, total_cnt2 / total_cnt))
                feat_vals_freq = feat_vals[feat] = []
                for val_id, cnt in item.items():
                    val = vals[val_id]
                    feat_vals_freq.append((val, cnt, cnt / total_cnt2))
                feat_vals_freq.sort(key=lambda x: (x[1], x[0]), reverse=True)
            feats_freq.sort(key=lambda x: (x[1], x[0]), reverse=True)

        self._most_probable_wform_feats = {}
        wtl_cnts = self._wform_tag_cnts
            # {wform_id: {tag_id: {lemma_id: cnt}}}
        for wform_id, item in self._wform_feat_cnts.items():  # {wform_id:
                         # {tag_id: {lemma_id: {feat_id: {val_id: cnt}}}}}
            tl_cnts = wtl_cnts[wform_id]  # {tag_id: {lemma_id: cnt}}
            feat_cnts = {}
            for tag_id, item in item.items():  # {tag_id:
                  # {lemma_id: {feat_id: {val_id: cnt}}}}
                l_cnts = tl_cnts[tag_id]  # {lemma_id: cnt}
                for lemma_id, item in item.items():  # {lemma_id: 
                                      # {feat_id: {val_id: cnt}}}
                    all_feats = set(tag_feats[tag_id])
                    _cnts = l_cnts[lemma_id]  # cnt
                    for feat_id, item in item.items():  # {feat_id: 
                                                   # {val_id: cnt}}
                        cnts = 0
                        all_feats.remove(feat_id)
                        for val_id, cnt in item.items():  # {val_id: cnt}
                            cnts += cnt
                            feat_cnts[feat_id][val_id] = \
                                feat_cnts.setdefault(feat_id, {}) \
                                         .get(val_id, 0) + cnt
                        if _cnts > cnts:
                            feat_cnts[feat_id][0] = \
                                feat_cnts.setdefault(feat_id, {}) \
                                         .get(0, 0) + _cnts - cnts
                    for feat_id in all_feats:
                        feat_cnts[feat_id][0] = \
                            feat_cnts.setdefault(feat_id, {}) \
                                     .get(0, 0) + _cnts
            for feat_id, val_cnts in feat_cnts.items():
                val_id, cnt = max(val_cnts.items(), key=itemgetter(1))
                n = sum(val_cnts.values())
                # Don't add rare wforms to the feat dictionary
                # Only add quite unambiguous wforms
                if n >= cnt_thresh and cnt / n >= ambiguity_thresh:
                    self._most_probable_wform_feats \
                        .setdefault(wform_id, {})[feat_id] = val_id

        self._most_probable_lemma_feats = {}
        lt_cnts = self._lemma_tag_cnts
            # {lemma_id: {tag_id: cnt}}
        for lemma_id, item in self._lemma_feat_cnts.items():  # {lemma_id:
                                     # {tag_id: {feat_id: {val_id: cnt}}}}
            t_cnts = lt_cnts[lemma_id]  # {tag_id: cnt}
            feat_cnts = {}
            for tag_id, item in item.items():  # {tag_id:
                              # {feat_id: {val_id: cnt}}}
                all_feats = set(tag_feats[tag_id])
                _cnts = t_cnts[tag_id]  # cnt
                for feat_id, item in item.items():  # {feat_id: 
                                               # {val_id: cnt}}
                    cnts = 0
                    all_feats.remove(feat_id)
                    for val_id, cnt in item.items():  # {val_id: cnt}
                        cnts += cnt
                        feat_cnts[feat_id][val_id] = \
                            feat_cnts.setdefault(feat_id, {}) \
                                     .get(val_id, 0) + cnt
                    if _cnts > cnts:
                        feat_cnts[feat_id][0] = \
                            feat_cnts.setdefault(feat_id, {}) \
                                     .get(0, 0) + _cnts - cnts
                for feat_id in all_feats:
                    feat_cnts[feat_id][0] = \
                        feat_cnts.setdefault(feat_id, {}) \
                                 .get(0, 0) + _cnts
            for feat_id, val_cnts in feat_cnts.items():
                val_id, cnt = max(val_cnts.items(), key=itemgetter(1))
                n = sum(val_cnts.values())
                # Don't add rare lemmata to the feat dictionary
                # Only add quite unambiguous lemmata
                if n >= cnt_thresh and cnt / n >= ambiguity_thresh:
                    self._most_probable_lemma_feats \
                        .setdefault(lemma_id, {})[feat_id] = val_id

        if log_file:
            print('done.', file=log_file)

    def get_tags(self):
        """Return a set of all known tag labels.
        
        :rtype: set(str)
        """
        return set(self._tags)
    
    def get_feats(self):
        """Return a dict of all known feats with their possible value labels.

        :rtype: dict(str: set(str))
        """
        return {feat: set(val for val in self._feat_vals[i])
                          for i, feat in enumerate(self._feats)}

    def get_tag_feats(self, tag):
        """Return a set of all known feat labels for a given tag.
        
        :rtype: set(str)
        """
        tag_id = self._tags_id.get(tag)
        assert tag_id is not None, 'ERROR: Unknown tag specified'
        return set(self._feat_vals[feat_id] \
                       for feat_id in self._tag_feats[tag_id])

    def get_tags_freq(self):
        """Return tags ordered by frequency.
        
        :rtype: [(tag, count, tag_freq)]
        """
        return self._tags_freq[:]
    
    def get_feats_freq(self, tag):
        """Return feats ordered by frequency for a given *tag*.
        
        :rtype: [(feat, count, feat_freq)]
        """
        return self._feats_freq[tag][:]
    
    def get_feat_vals_freq(self, tag, feat):
        """Return feat values ordered by frequency for a given *tag* and
        *feat*.
        
        :rtype: [(val, count, val_freq)]
        """
        return self._feat_vals_freq[tag].get(feat, [])[:]

    def wform_isknown(self, wform, tag=None):
        res = self._wforms_id.get(wform)
        if res is None:
            res = self._wforms_id.get(wform.lower())
        if res is None:
            res = self._lemmata_id.get(wform)
            if res is None:
                res = self._lemmata_id.get(wform.lower())
            if res is not None and tag:
                res = self._lemma_tag_cnts[res].get(self._tags_id[tag])
        elif tag:
            res = self._wform_tag_cnts[res].get(self._tags_id[tag])
        return res is not None

    def lemma_isknown(self, lemma, tag=None):
        res = self._lemmata_id.get(lemma)
        if res is None:
            res = self._lemmata_id.get(lemma.lower())
        if res is not None and tag:
            res = self._lemma_tag_cnts[res].get(self._tags_id[tag])
        return res is not None

    def most_common_tag(self):
        """Return most common tag for the whole corpus"""
        return self._most_common_tag

    def predict_tag(self, wform, isfirst=False, cnt_thresh=None):
        """If the *wform* has a trusted tag, then return that tag with
        a relevance coef equals to 1. Elsewise, we choose the most common tag
        for that *wform* and calculate some empirical value as a relevance
        coef.

        The relevance coef may be used to improve an alternative tag
        prediction algorithm.

        :param isfirst: pointed whether the *wform* is a first word in the sentence
        :param cnt_thresh: if the *wform* was met in the corpus less than
                           *cnt_thresh* times, then the relevance coef will be
                           discounted by (count / cnt_thresh)
        :type cnt_thresh: int
        :return: tag predicted and relevance coef
        :rtype: tuple(str, float)

        If the *wform* is not known then return (None, None)"""
        if cnt_thresh is None:
            cnt_thresh = self._cnt_thresh

        tag, coef = None, None
        wform_id = self._wforms_id.get(wform)
        wform_id2 = None
        if wform_id is None:
            wform_id = self._wforms_id.get(wform.lower())
        elif isfirst and wform.istitle():
            wform_id2 = self._wforms_id.get(wform.lower())
        if wform_id is not None:
            if wform_id2 is None:
                tag_id = self._most_probable_tags.get(wform_id)
                if tag_id is not None:
                    tag, coef = self._tags[tag_id], 1.
            if tag is None:
                tag_cnts = {
                    x: sum(y.values())
                        for x, y in self._wform_tag_cnts[wform_id].items()
                }
                if wform_id2 is not None:
                    for x, y in [
                        (x, sum(y.values()))
                            for x, y in self._wform_tag_cnts[wform_id2].items()
                    ]:
                        tag_cnts[x] = tag_cnts.get(x, 0) + y
                #n = sum(tag_cnts.values())
                # x[0] for stability
                tag_id, cnt = max(tag_cnts.items(), key=lambda x: (x[1], x[0]))
                del tag_cnts[tag_id]
                cnt2 = max(tag_cnts.values()) if tag_cnts else 0
                tag = self._tags[tag_id]
                coef = (cnt - cnt2) * min((cnt - cnt2) / cnt_thresh, 1.) / cnt
                #if isfirst and self._wforms[wform_id].istitle():
                #    tag2, coef2 = self.predict_tag(wform.lower())
                #    if coef2 and coef2 > coef:
                #        tag, coef = tag2, coef2
        return tag, coef

    def predict_lemma(self, wform, tag, isfirst=False, cnt_thresh=None):
        """Choose the most common lemma for the *wform* and *tag* and calculate
        some empirical value as a relevance coef.

        The relevance coef may be used to improve an alternative lemma
        prediction algorithm.

        :isfirst: True if the *wform* is first in a sequence
        :param cnt_thresh: if the *wform* was met in corpus less than
                           *cnt_thresh* times, then the relevance coef will be
                           discounted by (count / cnt_thresh)
        :type cnt_thresh: int
        :return: lemma chosen and relevance coef
        :rtype: tuple(str, float)
        """
        tag_id = self._tags_id.get(tag)
        assert tag_id is not None, 'ERROR: Unknown tag specified'

        if cnt_thresh is None:
            cnt_thresh = self._cnt_thresh

        lemma, coef = None, None
        if wform.isalpha():
            lemma_cnts = None
            wform_hascaps = not wform.islower()

            # Trying to find wform in a dict
            wform_id = self._wforms_id.get(wform)
            if wform_id is not None:
                lemma_cnts = self._wform_tag_cnts[wform_id].get(tag_id)
                if not lemma_cnts:
                    wform_id = None
            if wform_id is None and wform_hascaps:
                wform_id = self._wforms_id.get(wform.lower())
                if wform_id is not None:
                    lemma_cnts = self._wform_tag_cnts[wform_id].get(tag_id)

            # Wform is in a dict, so lemma is in a dict, too
            if lemma_cnts:
                lemma_cnts = sorted(lemma_cnts.items(),
                                    # x[0] for stability
                                    key=lambda x: (x[1], x[0]), reverse=True)
                lemma_id, cnt = lemma_cnts[0]
                cnt2 = lemma_cnts[1][1] if len(lemma_cnts) > 1 else 0
                lemma = self._lemmata[lemma_id]
                # Strong sorcery. Don't think about it.
                coef = (cnt - cnt2) * min((cnt - cnt2) / cnt_thresh, 1.) / cnt

            # Wform is not in a dict
            else:
                # Known endings for a given tag
                eds = self._common_endings.get(tag_id)
                if eds:
                    # Search first known lemma amongst generated ones sorted by
                    # frequency and length. In addition, keep in mind the first
                    # generated lemma.
                    known_lemma = gen_lemma = None
                    for i in range(len(wform)):
                        # Check all known lemma endings for a given wform
                        # ending
                        eds_ = eds.get(wform[i:])
                        if eds_:
                            # eds_ - list of tuples sorted by frequency
                            for ed, _ in eds_:
                                lemma_ = wform[:i] + ed
                                # If lemma is in a dict, we've found it.
                                lemma_id = self._lemmata_id.get(lemma_)
                                if lemma_id is not None \
                               and self._lemma_tag_cnts[lemma_id].get(tag_id):
                                    known_lemma = lemma_
                                    break
                                # Save the first generated lemma
                                if not gen_lemma:
                                    gen_lemma = lemma_
                            else:
                                continue
                            break
                    # If known lemma has not found and wform is not lower, do
                    # all the same for its lower form
                    if not known_lemma and wform_hascaps:
                        wform_lower = wform.lower()
                        for i in range(len(wform)):
                            eds_ = eds.get(wform_lower[i:])
                            if eds_:
                                for ed, _ in eds_:
                                    lemma_ = wform_lower[:i] + ed
                                    lemma_id = self._lemmata_id.get(lemma_)
                                    if lemma_id is not None \
                                   and self._lemma_tag_cnts[lemma_id].get(tag_id):
                                        known_lemma = lemma_
                                        break
                                    if not gen_lemma:
                                        gen_lemma = lemma_
                                        ## Has no effect:
                                        #gen_lemma = lemma_ \
                                        #    if not isfirst \
                                        #    or tag_id \
                                        #        not in self._capitalized_tags \
                                        #    else wform[:i] + ed
                                else:
                                    continue
                                break
                    if known_lemma:
                        lemma, coef = known_lemma, .9
                    elif gen_lemma:
                        lemma, coef = gen_lemma, 0

        # If lemma is not found after all, set lemma = wform
        if not lemma:
            lemma, coef = wform, 0
        return lemma, coef

    def predict_feat(self, feat, wform, lemma, tag, cnt_thresh=None):
        """Choose the most common *feat* for the *wform*, *tag*, and *lemma*
        and calculate some empirical value as a relevance coef.

        The relevance coef may be used to improve an alternative feats
        prediction algorithm.

        :param cnt_thresh: if the *wform* was met in corpus less than
                           *cnt_thresh* times, then the relevance coef will be
                           discounted by (count / cnt_thresh)
        :type cnt_thresh: int
        :return: feat predicted and relevance coef
        :rtype: tuple(str, float)

        If the *wform* is not known then return (None, None)"""
        feat_id = self._feats_id.get(feat)
        assert feat_id is not None, 'ERROR: Unknown feat specified'
        tag_id = self._tags_id.get(tag)
        assert tag_id is not None, 'ERROR: Unknown tag specified'

        if cnt_thresh is None:
            cnt_thresh = self._cnt_thresh

        val, coef = None, None
        if feat_id not in self._tag_feats[tag_id]:
            val, coef = None, 1.
        else:
            wform_id = self._wforms_id.get(wform)
            if wform_id is None:
                wform_id = self._wforms_id.get(wform.lower())
            if wform_id is not None:
                val_id = self._most_probable_wform_feats.get(wform_id, {}).get(
                    feat_id, self._most_probable_lemma_feats.get(
                        self._lemmata_id.get(lemma), {}
                    ).get(feat_id)
                )
                if val_id is not None:
                    val, coef = self._feat_vals[feat_id][val_id], 1.
                else:
                    feat_cnts = self._wform_feat_cnts[wform_id] \
                                    .get(tag_id, {}) \
                                    .get(self._lemmata_id.get(lemma))
                        # {wform_id: {tag_id: {lemma_id:
                        #    {feat_id: {val_id: cnt}}}}}
                    tag_cnt = sum(
                        self._wform_tag_cnts[wform_id].get(tag_id, {}).values()
                    )
                        # {wform_id: {tag_id:
                        #   {lemma_id: cnt}}}
                    if feat_cnts:
                        val_cnts = dict(feat_cnts.get(feat_id, {}))  # {feat_id:
                        #                                         {val_id: cnt}}
                        # add '_' with count = count(tag) - count(feat)
                        val_cnts.update({0: tag_cnt - sum(val_cnts.values())})
                        # x[0] for stability
                        val_id, cnt = max(
                            val_cnts.items(), key=lambda x: (x[1], x[0])
                        )
                        del val_cnts[val_id]
                        cnt2 = max(val_cnts.values()) if val_cnts else 0
                        val = self._feat_vals[feat_id][val_id]
                        coef = (cnt - cnt2) \
                             * min((cnt - cnt2) / cnt_thresh, 1.) \
                             / cnt
                    else:  # feat_cnts == {}
                        val = self._feat_vals[feat_id][0]  # '_'
                        coef = min(tag_cnt / cnt_thresh, 1.)
        return val, coef



# -*- coding: utf-8 -*-
# Corpuscula project: Utilities
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
from difflib import SequenceMatcher
import re
import os
import shutil
import sys
from urllib.request import urlopen

LOG_FILE=sys.stderr


def vote(sequence, weights=None):
    """Return a list of unique objects from the *sequence* sorted by frequency.

    :rtype: list[(obj, count, freq)]
    """
    cands = {}
    for o, w in zip(sequence, weights if weights else iter(lambda: 1, 2)):
        cands[o] = cands.get(o, 0) + w
    cnt = sum(cands.values())
    res = [(x, y, y / cnt) for x, y in cands.items()]
    res.sort(key=lambda x: (x[1], x[0]), reverse=True)
    return res

def find_affixes(wform, lemma, lower=False):
    """Find the longest common part of the given *wform* and *lemma*.

    :param lower: if True then return values will be always in lower case
    :return: prefix, common part, suffix/flexion of the *wform*;
             prefix, common part, suffix/flexion of the *lemma*
    :rtype: str, str, str, str, str, str
    """
    if lower:
        lex = wform = wform.lower()
        lem = lemma = lemma.lower()
    else:
        lex = wform.lower()
        lem = lemma.lower()
    lex = lex.replace('ё', 'е')
    lem = lem.replace('ё', 'е')
    a, b, size = SequenceMatcher(None, lex, lem, False) \
        .find_longest_match(0, len(lex), 0, len(lem))
    return wform[:a], wform[a:a + size], wform[a + size:], \
           lemma[:b], lemma[b:b + size], lemma[b + size:]
    #       lemma[:b], wform[a:a + size], lemma[b + size:]

def find_file(prefix, ext=None, dname=None):
    """Return a full name of the file with specified *prefix* if it's present
    in the directory *dname*. If there are several such files, a name of the
    first one will be returned.
    
    :param ext: an extension of the target file
    :param dname: a name of the directory for searching. If None, then the
                  current directory will be used
    """
    for fname in os.listdir(dname):
        if fname.startswith(prefix) and (not ext or fname.endswith('.' + ext)):
            full_fname = os.path.join(dname if dname is not None else '',
                                      fname)
            if os.path.isfile(full_fname):
                res = full_fname
                break
    else:
        res = None
    return res

def rmdir(dname):
    """Remove a directory *dname*"""
    for fname in os.listdir(dname):
        path = os.path.join(dname, fname)
        if os.path.isdir(path):
            rmdir(path)
        else:
            os.remove(path)
    os.rmdir(dname)

def print_progress(current_value, end_value=10, step=1, start_value=0,
                   max_width=60, file=LOG_FILE):
    """Show progress indicator.

    :param end_value: None means unknown; 0 means show finish
    """
    # TODO: optimize all this trash
    cur_value = current_value
    isfinish = False
    if end_value == 0:
        end_value = None
        isfinish = True
    if end_value is not None:
        if start_value:
            cur_value -= start_value
            end_value -= start_value
        end_value /= step
        pend_value = int(end_value) + int(not end_value.is_integer())
    cur_value /= step
    pcur_value = int(cur_value) + int(not cur_value.is_integer())

    full_lines = (pcur_value - 1) // max_width if pcur_value else 0
    if pcur_value > 0:
        pcur_value = pcur_value % max_width
        if pcur_value == 0:
            pcur_value = max_width

    if end_value is not None:
        total_lines = pend_value // max_width + (pend_value % max_width != 0)
        pend_value = min(max_width, pend_value - full_lines * max_width)
        print('\r{}{}{}{}{}{}'.format(
            '>' if full_lines else '[',
            '#' * pcur_value,
            (' ' * (pend_value - pcur_value)),
            ']' if full_lines + 1 >= total_lines else '>',
            ' {:.0f}%'.format(100 * cur_value / end_value),
            ' ' * (max_width - pend_value)
        ), end='' if cur_value < end_value else '\n', file=file)
    else:
        pcur_value += isfinish - 1
        print('\r{}{}{} {}{}'.format(
            '>' if full_lines else '[',
            '=' * pcur_value,
            ']' if isfinish else '>',
            current_value,
            ' ' * (max_width - pcur_value)
        ), end='\n' if isfinish else '', file=file)

COPY_BUF_SIZE = 16 * 1024
CALLBACK_CHUNK_SIZE = 1024 * 1024
def copyfileobj(fsrc, fdst, buf_size=COPY_BUF_SIZE,
                callback=None, callback_chunk_size=CALLBACK_CHUNK_SIZE):
    """Copy data from file-like object *fsrc* to file-like object *fdst*
    (extends shutil.copyfileobj)

    :param fsrc: input stream
    :param fdst: output stream
    :param buf_size: buffer size (in bytes) for r/w operations
    :type buf_size: int
    :param callback: callback function
    :type callback: function(bytes_read, chunks_read, last_chunk_size)
    :param callback_chunk_size: invoke *callback* after every
                                *callback_chunk_size* bytes read
    :type callback_chunk_count: int
    """
    bytes_read = 0
    chunks_read = 0
    chunk_size = 0
    def run_callback_if_need(buf=None):
        nonlocal bytes_read, chunks_read, chunk_size
        urgent_need = False
        if callback:
            if buf:
                buf_len = len(buf)
                bytes_read += buf_len
                chunk_size += buf_len
            elif chunk_size != 0:
                urgent_need = True
            if chunk_size >= callback_chunk_size or urgent_need:
                chunks_read += 1
                callback(bytes_read, chunks_read, chunk_size)
                chunk_size = 0
    while True:
        buf = fsrc.read(buf_size)
        if not buf:
            break
        fdst.write(buf)
        run_callback_if_need(buf)
    run_callback_if_need()

def copy_file(src, dst, buf_size=COPY_BUF_SIZE, **kwargs):
    """Copy data from file *src* to file *dst*.

    :param src: source file name
    :param dst: destination file name
    :param **kwargs: keyword params for ``copyfileobj()`` method
    """
    with open(src, 'rb') as fsrc, open(dst, 'wb') as fdst:
        copyfileobj(fsrc, fdst, **kwargs)

DIR_ACCESS_RIGHTS = 0o755
def download_file(url, dpath=None, fname=None, chunk_size=CALLBACK_CHUNK_SIZE,
                  file_noless=None, overwrite=True, log_msg=None,
                  silent=False):
    """Download file from *url* with progress indicator

    :param url: source url
    :type url: str
    :param dpath: path to the destination directory;
                  if None, then the current work directory will be used
    :type dpath: str
    :param fname: result file name;
                  if None then the name from url will be kept
    :type fname: str
    :param chunk_size: show progress after every *chunk_size* bytes read
    :type chunk_size: int
    :file_noless: if the file is smaller, then don't download it and keep
                  already downloaded one (if exists)
    :type file_noless: int
    :param overwrite: if False and the file is exist, overwrite it
    :type overwrite: bool
    :param log_msg: message that will be printed before downloading
    :type log_msg: str
    :param silent: do not show progress
    :type silent: bool

    :return: path to the downloaded file
    :rtype: str
    """
    if dpath is None:
        dpath = '.'
    if dpath and not os.path.exists(dpath):
        os.makedirs(dpath, DIR_ACCESS_RIGHTS)
    fpath = os.path.join(dpath, re.sub('^.+/', '', url) if fname is None else
                                fname)
    if overwrite or not os.path.isfile(fpath):
        if not silent and log_msg is not None:
            print(log_msg, file=LOG_FILE)
        fpath_ = fpath + '$'
        with urlopen(url) as f_in:
            f_in_length = f_in.length
            if file_noless is not None:
                if f_in_length is None:
                    raise RuntimeError(
                        'Perhaps the file was removed from the server. '
                        'Check the url manually:\n' + url
                    )
                elif f_in_length < file_noless:
                    raise RuntimeError(
                        'The size of the downloading file less than threshold '
                        '({} bytes vs. {}). '
                            .format(f_in_length, file_noless) +
                        'Check the url manually:\n' + url
                    )
            total_, used_, free_ = shutil.disk_usage(dpath)
            if f_in_length is not None and f_in_length > free_:
                 raise RuntimeError('Not enough space in {} to download file'
                                        .format(dpath))
            chunks_count = f_in_length / CALLBACK_CHUNK_SIZE \
                if f_in_length is not None else None
            if chunks_count is not None:
                chunks_count = int(chunks_count) \
                             + (not chunks_count.is_integer())
            prev_line_no = 0
            if not silent and (chunks_count is None or chunks_count > 2):
                print_progress(0, chunks_count)
            bytes = 0
            try:
                def callback(bytes_read, chunks_read, last_chunk_size):
                    nonlocal bytes
                    bytes = bytes_read
                    prev_line_no = print_progress(
                        bytes_read if chunks_count is None else chunks_read,  # current
                        chunks_count,                                         # end
                        CALLBACK_CHUNK_SIZE if chunks_count is None else 1    # step
                    ) if not silent and (chunks_count is None
                                      or chunks_count > 2) else None
                with open(fpath_, 'wb') as f_out:
                    copyfileobj(
                        f_in, f_out, 
                        callback=callback
                                     if not silent and (chunks_count is None
                                                     or chunks_count > 2) else
                                 None
                    )
            except KeyboardInterrupt as e:
                try:
                    os.remove(fpath_)
                except OSError:
                    pass
                raise e
            if not silent and chunks_count is None:
                print_progress(bytes, 0, CALLBACK_CHUNK_SIZE)
        os.replace(fpath_, fpath)
        if not silent:
            print('done: {} bytes'.format(os.path.getsize(fpath)),
                  file=LOG_FILE)
    return fpath

def read_bz2(apath, encoding='utf-8', errors='ignore', process_line=None):
    """Read lines from a file in bz2 archive.

    :param process_line: a function that will be invoked to process each file's
                         line. If its result is a list, then it will be
                         returned by lines
    :type process_line: callable
    """
    from bz2 import open as bz2_open
    with bz2_open(apath, mode='rt', encoding=encoding, errors=errors) as f:
        for line in f:
            if process_line:
                line = process_line(line)
                if isinstance(line, list):
                    for line_ in line:
                        yield line_
                    continue
            yield line

def read_rar(apath, fname, encoding='utf-8', errors='ignore',
             process_line=None):
    """Read lines from a file in rar archive.

    :param fname: a name of the file in the archive
    :param process_line: a function that will be invoked to process each file's
                         line. If its result is a list, then it will be
                         returned by lines
    :type process_line: callable
    """
    from rarfile import RarFile
    with RarFile(apath) as a:
        with a.open(fname) as f:
            for line in f:
                line = line.decode(encoding=encoding, errors=errors)
                if process_line:
                    line = process_line(line)
                    if isinstance(line, list):
                        for line_ in line:
                            yield line_
                        continue
                yield line

def read_zip(apath, fname, encoding='utf-8', errors='ignore',
             process_line=None):
    """Read lines from a file in zip archive.

    :param fname: a name of the file in the archive
    :param process_line: a function that will be invoked to process each file's
                         line. If its result is a list, then it will be
                         returned by lines
    :type process_line: callable
    """
    from zipfile import ZipFile
    with ZipFile(apath) as a:
        with a.open(fname) as f:
            for line in f:
                line = line.decode(encoding=encoding, errors=errors)
                if process_line:
                    line = process_line(line)
                    if isinstance(line, list):
                        for line_ in line:
                            yield line_
                        continue
                yield line



# -*- coding: utf-8 -*-
# Corpuscula project: Wrapper for Wikipedia
#
# Copyright (C) 2019-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Tools for downloading and converting Russian part of Wikipedia.
Includes wrapper to simplify the further processing.
"""
from html import unescape
from re import compile as re_compile, match as re_match, sub as re_sub

from corpuscula.corpus_utils import _AbstractCorpus, download_corpus, \
                                    remove_corpus, get_corpus_fpath
from corpuscula.utils import LOG_FILE, print_progress, read_bz2


WIKIPEDIA_RU = 'Wikipedia.RU'
WIKIPEDIA_RU_URL = 'https://dumps.wikimedia.org/ruwiki/latest/ruwiki-latest-pages-articles.xml.bz2'
WIKIPEDIA_RU_DNAME = 'wikipedia_ru'
def _consts(lang='RU'):
    if lang == 'RU':
        res = {'name' : WIKIPEDIA_RU,
               'url'  : WIKIPEDIA_RU_URL,
               'dname': WIKIPEDIA_RU_DNAME}
    else:
        raise ValueError('ERROR: Lang "{}" is not supported yet'.format(lang))
    return res

def download_wikipedia(lang='RU', root_dir=None, overwrite=True):
    """Downloaded Wikipedia dump.
    
    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    :param overwrite: False means do not download the corpus if it's already
                      kept in the corpus storage
    """
    consts = _consts(lang=lang)
    return download_corpus(consts['name'], consts['url'],
                           dname=consts['dname'], root_dir=root_dir,
                           overwrite=overwrite, file_noless=3000000000)

def remove_wikipedia(lang='RU', root_dir=None):
    """Remove Wikipedia dump.

    :param root_dir: path to the root storage. If None, default value will
                     be used
    :type root_dir: str
    """
    remove_corpus(_consts(lang=lang)['dname'], root_dir=root_dir)

def _get_titles(fpath, silent=False):
    return _get_all(fpath, what='titles', silent=silent)

def _get_articles(fpath, silent=False):
    return _get_all(fpath, what='articles', silent=silent)

def _get_templates(fpath, silent=False):
    return _get_all(fpath, what='templates', silent=silent)

re_html = re_compile(r'<[^<>]+?>')
re_spenter = re_compile(r'[^\S\n]+\n')
def _get_all(fpath, what=None, silent=False):

    def read_txt():
        with open(fpath, 'rt', encoding='utf-8-sig') as f:
            for line in f:
                yield line

    if not silent:
        print('Process Wikipedia', file=LOG_FILE)
    id_ = title = text = None
    isobject = istable = issquare = 0
    iscomment = False
    inside_tag, tag_cnt = None, 0
    ready_for_save = False
    enters = 0
    istemplate = False
    namespaces = set()

    article_no = -1
    for line_no, line in enumerate(
        read_bz2(fpath) if fpath[-4:].lower() == '.bz2' else read_txt()
    ):

        if text is None:
            line = line.strip()

            ns = re_match(r'<namespace \S+ case="first-letter">'
                          r'([^<]+)</namespace>', line)
            if ns:
                ns, = ns.groups()
                namespaces.add(ns)
                continue

            if line.startswith('<title>'):
                title = re_html.sub('', line)
                pos = title.find(':')
                if pos > 0:
                    ns = title[:pos]
                    if ns in namespaces:
                        if (not what or what == 'templates') and ns == 'Шаблон':
                            istemplate = True
                            title = title[pos + 1:]
                        else:
                            title = None
                            continue
                article_no += 1
                if not silent:
                    print_progress(article_no, end_value=None, step=100000,
                                   file=LOG_FILE)
                id_ = None

            elif not id_ and line.startswith('<id>'):
                id_ = re_html.sub('', line)
                if what == 'titles':
                    yield id_, title.strip()

            elif what != 'titles' and title and line.startswith('<text '):
                text = ''

        line_isempty = False
        if text is not None:
            line = line.rstrip()
            if line == '':
                line_isempty = True
            if line.endswith('</text>'):
                ready_for_save = True
            line = re_html.sub('', line)

            if istemplate:
                text += line
                if ready_for_save:
                    text = re_spenter.sub('\n', text).strip()
                    yield (id_, title.strip(), text) if what else \
                          (id_, title.strip(), None, text)
                    id_ = title = text = None
                    istemplate = False
                    ready_for_save = False
                else:
                    text += '\n'
                continue
            if what and what != 'articles':
                continue

            continue_processing = True
            line = unescape(line)
            line_ = None
            while line_ != line:
                if line_ is not None:
                    line = line_
                line_ = re_sub(r'<(\w+)(?:\s[^>]*)?>.*?</\1>', '', line)

            if inside_tag:
                pos = line.rfind('</' + inside_tag + '>')
                if pos >= 0:
                    line = line[pos + 3 + len(inside_tag):]
                    tag_cnt -= 1
                    if not tag_cnt:
                        inside_tag = None
                else:
                    continue_processing = False

            if continue_processing:
                line = re_sub('<[^>]+/>', '', line)
                for token in ('strike', 'math', 'nowiki', 'ref', 'gallery',
                              'imagemap', 'div'):
                    pos = line.find('<' + token)
                    if pos >= 0:
                        line = line[:pos]
                        inside_tag = token
                        tag_cnt += 1
                        break

                line = line.replace('́', '').replace('()', '')
                line = re_html.sub(' ', line)
                if '-->' in line:
                    iscomment = False
                if iscomment:
                    continue
                if '<!--' in line:
                    iscomment = True
                line = re_sub(r'<!--.*', '', line)
                line = re_sub(r'.*-->', '', line)

                for _ in range(1):

                    if line == '----':
                        line = ''
                        break

                    def process_cu(match):
                        head, res_, tail = match.groups()
                        if '[[' in res_ or ']]' in res_:
                            res = ''
                        else:
                            res = re_sub(r'^(?:'
                                r'СС2'
                            r')\|([^|]+)\|([^|]+)\|([^|]+)(?:\|.*)?$',
                            '\g<1> \g<2> \g<3>', res_)
                            if res == res_:
                                res = re_sub(r'^(?:'
                                    r'num|число'
                                r')\|([^|]+)\|([^|]+)(?:\|.*)?$',
                                '\g<1> \g<2>', res_)
                                if res == res_:
                                    res = re_sub(r'^(?:'
                                        r'num|число|lang-\w+|nobr|вьетнамго|'
                                        r'(?:год|year)[^|]*'
                                    r')\|([^|]+)(?:\|.*)?$', '\g<1>', res_)
                                    if res == res_:
                                        res = ''
                        return head + re_sub(r"''+", "", res) + tail

                    line_ = None
                    while line_ != line:
                        if line_ is not None:
                            line = line_
                        line_ = re_sub(r'''(?x)
                            ( (?:^|[^{]) (?:\{\{)* )
                            \{\{
                            (
                                (?:
                                    [{}]? [^{}]
                                )*
                                [{}]?
                            )
                            \}\}
                            ( (?:\}\})* (?:[^}]|$) )
                        ''', process_cu, line)
                    if isobject:
                        cnt = line.count('}}')
                        if cnt > 0:
                            pos = line.rfind('}}')
                            if pos >= 0:
                                line = line[pos + 2:]
                                isobject -= cnt
                    else:
                        line.replace('}}', '')  # just in case

                    line_ = None
                    while line_ != line:
                        if line_ is not None:
                            line = line_
                        line_ = re_sub(r'''(?x)
                            \{\|
                            (
                                (?:
                                    [{}]? [^{}]
                                )*
                                [{}]?
                            )
                            \|\}
                        ''', '', line)
                    if istable:
                        cnt = line.count('|}')
                        if cnt > 0:
                            pos = line.rfind('|}')
                            if pos >= 0:
                                line_ = line[pos + 2:]
                                if not line_ or line_[0] != '}':
                                    line = line_
                                    istable -= cnt
                    else:
                        line.replace('|}', '')  # just in case

                    def process_sq(match):
                        redir, head, res_, tail = match.groups()
                        if redir:
                            res = ''
                        else:
                            res = re_sub(r'(?i)^:?(?:'
                                r'File|Файл|Media|Медиа|Category|Категория'
                            r')\:.+$', '', res_)
                            if res == res_:
                                # [[Москва (город)|Москве]] -> Москве
                                res = re_sub(r'^.+?\|([^|]+?)(?:\|.*)?$', '\g<1>', res_)
                                # [[царство (в биологии)|]] -> царство
                                if res == res_:
                                    res = re_sub(r'^([^|(]+?)\s*\(.+?\)\s*\|\s*',
                                                 '\g<1>', res_)
                                    if res == res_:
                                        # [[Викизнание: Новости|]] -> Новости
                                        res = re_sub(r'^.+?\:([^|:]+)\|\s*',
                                                     '\g<1>', res_)
                                        if res == res_:
                                            # [[sociowiki:Главная Страница]]
                                            # [[:Категория:Газеты Литвы]]
                                            res = re_sub(r'^\:?\S+?\:.+',
                                                         '', res_)
                                            #if res == res_:
                                            #    # [[тест]] -> тест
                                            #    pass
                        return head + re_sub(r"''+", "", res) + tail

                    line_ = None
                    while line_ != line:
                        if line_ is not None:
                            line = line_
                        line_ = re_sub(r'''(?xi)
                            (\#(?:redirect|перенаправление)\s*)?
                            ((?:^|[^\[])(?:\[\[)*)\[\[
                            (
                                (?:
                                    [\[\]]? [^\[\]]
                                )*
                                [\[\]]?
                            )
                            \]\]((?:\]\])*(?:[^\]]|$))
                        ''', process_sq, line)
                    if issquare:
                        cnt = line.count(']]')
                        if cnt > 0:
                            pos = line.rfind(']]')
                            if pos >= 0:
                                line = line[pos + 2:]
                                issquare -= cnt
                    else:
                        line.replace(']]', '')  # just in case

                    isobject_, istable_, issquare_ = isobject, istable, issquare
                    if not (istable or issquare):
                        cnt = line.count('{{')
                        if cnt > 0:
                            pos = line.find('{{')
                            if pos >= 0:
                                line = line[:pos]
                                isobject += cnt
                    if not (isobject or issquare):
                        cnt = line.count('{|')
                        if cnt > 0:
                            pos = line.find('{|')
                            if pos >= 0:
                                line = line[:pos]
                                istable += cnt
                    if not (isobject or istable):
                        cnt = line.count('[[')
                        if cnt > 0:
                            pos = line.find('[[')
                            if pos >= 0:
                                line = line[:pos]
                                issquare += cnt
                    if isobject_ or istable_ or issquare_:
                        continue

                    line_ = line = re_sub(r'^[*#:]+', '', line)
                    line = re_sub(r'^;([^:]*)\:', '\g<1>\n', line)
                    line = re_sub(r'^;([^:]*)', '\g<1>:', line)
                    #line = re_sub(r'^;', '', line)
                    line = re_sub(r'^==+', '', line)
                    line = re_sub(r'\s*==+$', '\n', line)
                    line = re_sub(r'\~~~+', '', line)

                    # [http://freebsd.org Сайт FreeBSD] -> Сайт FreeBSD
                    line = re_sub(r'\[\w+\://\S+\s+([^\[\]]+)\]', ' \g<1> ', line)
                    # [http://freebsd.org/] -> http://freebsd.org/
                    line = re_sub(r'\[(\w+\://[^\s\[\]]+)\]', ' \g<1> ', line)
                    # [mailto:name@example.com name@example.com] -> name@example.com
                    line = re_sub('\[mailto\:[^]]+?([^] ]+)\]', ' \g<1> ', line)
                    line = re_sub(r"''+", "", line)
                    break
                else:
                    continue_processing = False

                if continue_processing:
                    line = unescape(line)
                    line = re_sub(r'[ \t]+', ' ', line)
                    line = line.lstrip()
                    if line != '':
                        enters = line[-1] == '\n'
                    text += line
                    if enters < 2 and not ready_for_save:
                        if line != '' or line_isempty:
                            text += '\n'
                            enters += 1

        if ready_for_save:
            if text and enters:
                text = re_spenter.sub('\n', text[:-enters]).strip()
            yield (id_, title.strip(), text) if what else \
                  (id_, title.strip(), text, None)
            id_ = title = text = None
            isobject = istable = issquare = 0
            iscomment = False
            inside_tag, tag_cnt = None, 0
            ready_for_save = False
            enters = 0

    if article_no >= 0:
        if text:
            text = re_spenter.sub('\n', text).strip()
            yield (id_, title.strip(), text) if what else \
                  (id_, title.strip(), 
                   None if istemplate else text,
                   text if istemplate else None)
            article_no += 1
        article_no += 1
        if not silent:
            print_progress(article_no, end_value=0, step=100000, file=LOG_FILE)
            print('Wikipedia has been processed: {} lines, {} articles'
                      .format(line_no + 1, article_no),
                  file=LOG_FILE)


class Wikipedia(_AbstractCorpus):
    """Wrapper for Wikipedia corpus"""

    name = 'Wikipedia'
    _dl_name = 'download_wikipedia'

    def __init__(self, lang='RU', fpath=None, silent=False):
        self.name = _consts(lang=lang)['name']
        self._lang = lang
        if fpath:
            self._fpath = fpath
        self._silent = silent

    def _get_fpath(self):
        consts = _consts(lang=self._lang)
        fpath = self._fpath if hasattr(self, '_fpath') else \
                get_corpus_fpath(dname=consts['dname'], url=consts['url'])
        self.isfile(fpath)
        return fpath

    def titles(self, silent=None):
        return _get_titles(self._get_fpath(),
                           self._silent if silent is None else silent)

    def articles(self, silent=None):
        return _get_articles(self._get_fpath(),
                             self._silent if silent is None else silent)

    def templates(self, silent=None):
        return _get_templates(self._get_fpath(),
                              self._silent if silent is None else silent)
